=== modified file 'bzrlib/remote.py'
--- bzrlib/remote.py	2007-10-17 09:39:41 +0000
+++ bzrlib/remote.py	2007-10-25 01:24:51 +0000
@@ -492,26 +492,32 @@
             raise errors.UnexpectedSmartServerResponse(response)
 
     def unlock(self):
-        if self._lock_count == 1 and self._lock_mode == 'w':
-            # don't unlock if inside a write group.
-            if self.is_in_write_group():
-                raise errors.BzrError(
-                    'Must end write groups before releasing write locks.')
         self._lock_count -= 1
-        if not self._lock_count:
-            mode = self._lock_mode
-            self._lock_mode = None
+        if self._lock_count > 0:
+            return
+        old_mode = self._lock_mode
+        self._lock_mode = None
+        try:
+            # The real repository is responsible at present for raising an
+            # exception if it's in an unfinished write group.  However, it
+            # normally will *not* actually remove the lock from disk - that's
+            # done by the server on receiving the Repository.unlock call.
+            # This is just to let the _real_repository stay up to date.
             if self._real_repository is not None:
                 self._real_repository.unlock()
-            if mode != 'w':
+        finally:
+            # The rpc-level lock should be released even if there was a
+            # problem releasing the vfs-based lock.
+            if old_mode == 'w':
                 # Only write-locked repositories need to make a remote method
                 # call to perfom the unlock.
-                return
-            assert self._lock_token, 'Locked, but no token!'
-            token = self._lock_token
-            self._lock_token = None
-            if not self._leave_lock:
-                self._unlock(token)
+                assert self._lock_token, \
+                    '%s is locked, but has no token' \
+                    % self
+                old_token = self._lock_token
+                self._lock_token = None
+                if not self._leave_lock:
+                    self._unlock(old_token)
 
     def break_lock(self):
         # should hand off to the network

=== modified file 'bzrlib/repofmt/pack_repo.py'
--- bzrlib/repofmt/pack_repo.py	2007-10-24 05:17:39 +0000
+++ bzrlib/repofmt/pack_repo.py	2007-10-25 01:24:51 +0000
@@ -1534,8 +1534,12 @@
 
     def unlock(self):
         if self._write_lock_count == 1 and self._write_group is not None:
+            self.abort_write_group()
+            self._transaction = None
+            self._write_lock_count = 0
             raise errors.BzrError(
-                'Must end write groups before releasing write locks.')
+                'Must end write group before releasing write lock on %s'
+                % self)
         if self._write_lock_count:
             self._write_lock_count -= 1
             if not self._write_lock_count:

=== modified file 'bzrlib/repository.py'
--- bzrlib/repository.py	2007-10-24 05:59:51 +0000
+++ bzrlib/repository.py	2007-10-25 01:24:51 +0000
@@ -885,6 +885,8 @@
         if (self.control_files._lock_count == 1 and
             self.control_files._lock_mode == 'w'):
             if self._write_group is not None:
+                self.abort_write_group()
+                self.control_files.unlock()
                 raise errors.BzrError(
                     'Must end write groups before releasing write locks.')
         self.control_files.unlock()

=== modified file 'bzrlib/tests/repository_implementations/test_write_group.py'
--- bzrlib/tests/repository_implementations/test_write_group.py	2007-10-24 07:36:52 +0000
+++ bzrlib/tests/repository_implementations/test_write_group.py	2007-10-25 01:24:51 +0000
@@ -69,9 +69,11 @@
         # really to be sure it's used right, not for signalling
         # semantic information.
         self.assertRaises(errors.BzrError, repo.unlock)
-        self.assertTrue(repo.is_locked())
-        repo.commit_write_group()
-        repo.unlock()
+        # after this error occurs, the repository is unlocked, and the write
+        # group is gone.  you've had your chance, and you blew it. ;-)
+        self.assertFalse(repo.is_locked())
+        self.assertRaises(errors.BzrError, repo.commit_write_group)
+        self.assertRaises(errors.BzrError, repo.unlock)
 
     def test_is_in_write_group(self):
         repo = self.make_repository('.')

